// RayTracer.h: Interface for the RayTracer class.
//				This class ties everything together.  It allows for
//				the addition of objects and lights to the scene. It
//				casts and tracks the rays and generates the final
//				array of pixels to display on screen.
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_RAYTRACER_H__38A269FA_8098_47AB_90E8_71B6C4E7077F__INCLUDED_)
#define AFX_RAYTRACER_H__38A269FA_8098_47AB_90E8_71B6C4E7077F__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Vec4.hpp"	
#include "Vec3.hpp"	
#include "Colour.h"	
#include "Scene.h"
#include "Light.h"

#ifdef WIN32
#include <windows.h>
#endif

#ifndef WIN32
#define byte unsigned char
#endif

class RayTracer  
{
public:
	// array of pixels to display the image
	byte *img;

	// Window properties
	double xMin;
	double xMax;
	double yMin;
	double yMax;
	int Height;
	int Width;

	Scene ObjList;				// List of objects
	Colour BackgroundColor;		// background colour
	
	int Depth;					// Recursion depth
    double Kd;					// diffuse scalar
    double Ks;					// specular scalar
    double Kt;					// transparency scalar
    
    LightList Lights;			// List of lights for the scene

	// Constructor & Destructor
	RayTracer();
	virtual ~RayTracer();

	// Surface Properties
	Colour SurfaceAmbient;
	Colour SurfaceDiffuse;
	Colour SurfaceSpecular;
	double SurfaceReflective;
	double SurfaceTransparent;
	double SurfaceRefractionIndex;
    double SurfaceSpecExponent;
	
	// Tracks the rays progress through the scene.
	// Spawns additional rays for reflection and refraction
	Colour TraceRay(Ray r, int depth, const int trans);

	// Casts the rays and passes them to TraceRay.
	// Generates the image based on colour values 
	// returned from TraceRay
	void CastRays();

	// methods to add objects to the scene
	void Polygon(int n, Vec4 *P);
	void Cylinder(double r, double height);
	void Cone(double r, double height);
	void Sphere(double radius);

	// methods to transform the last object added to the scene
	void scale(double sx, double sy, double sz);
	void translate(double tx, double ty, double tz);
	void rotateX(double angle);
	void rotateY(double angle);
	void rotateZ(double angle);
	
	// methods to easily set the surface properties
	// of objects
	void Specular(double r, double g, double b);
    void SpecExponent(double pow);
	void Index(double index);
	void Transparent(int trans);
	void Reflective(double reflect);
	void Diffuse(double r, double g, double b);
	void Ambient(double r, double g, double b);

	// methods to add lights to the scene
	void AddLight(double intensity, double x, double y, double z);
	void AddLight(double intensity, double x, double y, double z, double R);
	
	// sets the background colour
	void Background(double r, double g, double b);

	// test function.  contains multiple scene definitons
	// which can be chosen by specifying arg. d is the 
	// recursion depth you would like to use.
    void Test(int arg, int d);

private:

};

#endif // !defined(AFX_RAYTRACER_H__38A269FA_8098_47AB_90E8_71B6C4E7077F__INCLUDED_)

