// Scene.cpp: Implementation of the Scene class.
//			This class defines a scene for our ray tracer.  It contains
//			the list of objects in the scene and operations to 
//			get properties of those objects and transform the objects.
//			It also has a method to calculate intersections and shading.
//////////////////////////////////////////////////////////////////////
#include <assert.h>
#include <stdlib.h>

#include "Scene.h"
#include "Sphere.h"
#include "Polygon.h"
#include "Cylinder.h"
#include "Cone.h"
#include "Misc.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Scene::Scene()
{
	NumObjects = 0;
	Item = new Object*[MAX_OBJECT];
}

Scene::~Scene()
{
	int i;
	for( i=0; i<NumObjects; i++ )
    {
		delete Item[i];
    }
}

//////////////////////////////////////////////////////////////////////
// Intersects : checks for intersection with all the objects in the 
// scene and the passed in ray.  returns true if at least one hit 
// occurred and which gets the closest object, the hitobject ho is
// built.
//////////////////////////////////////////////////////////////////////
bool Scene::Intersects(Ray ray, HitObject &ho, int& which, const int trans) const
{
	int i;
	double min_t = 1.e20;
	HitObject o;
    
	for( i=0; i<NumObjects; i++ )
	{
		if( Item[i]->Intersect(ray, o) && o.t < min_t )
		{
            if( trans != i )
			{
				min_t = o.t;	
			
				ho.Set(o); 
				which = i;
			}
        }
	}
    
    if ( min_t < 1.e20 )
		return true;
    else
		return false;
}

//////////////////////////////////////////////////////////////////////
// Methods to add objects to the scene.  These include:
// Polygon, Cylinder, Cone and Sphere
//////////////////////////////////////////////////////////////////////
bool Scene::AddPolygon(int n, Vec4 *P, SurfaceModel surface)
{
	if( NumObjects == MAX_OBJECT-1 )
		return false;

	Item[NumObjects] = new Polygon(n, P);
	Item[NumObjects]->SetSurface(surface);
	Type[NumObjects] = POLYGON;
    
	NumObjects++;
	return true;
}


bool Scene::AddCylinder(double r, double height, SurfaceModel surface)
{
	if( NumObjects == MAX_OBJECT-1 )
		return false;

	Item[NumObjects] = new Cylinder(r, height);
	Item[NumObjects]->SetSurface(surface);
	Type[NumObjects] = CYLINDER;
    
	NumObjects++;
	return true;
}

bool Scene::AddCone(double r, double height, SurfaceModel surface)
{
	if( NumObjects == MAX_OBJECT-1 )
		return false;

	Item[NumObjects] = new Cone(r, height);
	Item[NumObjects]->SetSurface(surface);
	Type[NumObjects] = CONE;
    
	NumObjects++;
	return true;
}

bool Scene::AddSphere(double r, SurfaceModel surface)
{
	if( NumObjects == MAX_OBJECT-1 )
		return false;

	Item[NumObjects] = new Sphere(0, 0, 0, r);
	Item[NumObjects]->SetSurface(surface);
	Type[NumObjects] = SPHERE;
    
	NumObjects++;
	return true;
}

//////////////////////////////////////////////////////////////////////
// Shade : Calculates the shading of an object specified by which.
// Based on the shade algorithm in the class textbook.
//////////////////////////////////////////////////////////////////////
Colour Scene::Shade(const Ray R, const int which, const HitObject ho, Light light) const
{
	Colour cd(0,0,0);
    Vec3 viewDir, lightDir, Normal;
	double scale;
	double mDotH, phong;
   
    Ray lightRay;
    HitObject lightho;
    int lightWhich;
	
	viewDir = -R.v3Dir;
	viewDir.Normalize();

	// direction of light
	lightDir.Set( (light.position - ho.hit).x,
				  (light.position - ho.hit).y,
				  (light.position - ho.hit).z );

    lightDir.Normalize();

	lightRay.Set(ho.hit, lightDir);
	
	// Transform Normal:  
	// N = M inverse Transpose * surfaceNormal
   	Normal = ho.TransformNormal(Item[which]->transMatrix);
	Normal.Normalize();

    // point light source
    if( light.radius == 0 )
    {
        if( !Intersects(lightRay, lightho, lightWhich, which ) )
		{		
            // diffuse color
	        scale = lightDir*Normal;

            if( scale < 0 ) 
				scale = 0;
            
	        if( scale > 1 )
                scale = 1;

			cd += scale * light.intensity * Item[which]->GetDiffuse();

			// Add phong specular component

			Vec3 half3 = viewDir + lightDir;
			half3.Normalize();

			mDotH = half3 * Normal;

			if (mDotH <= 0.0f)
			{
				return cd;
			}
			else
			{
				phong = pow(mDotH, Item[which]->GetSpecExponent());
				cd += light.intensity * phong * Item[which]->GetSpecular();
			}            
        }
    }

	return cd;	
}

//////////////////////////////////////////////////////////////////////
// Methods to get the surface properties of the object specified
// by the which variable.
//////////////////////////////////////////////////////////////////////
Colour Scene::GetAmbient(const int which) const
{
    assert(which<NumObjects);
    return Item[which]->GetAmbient();
}

double Scene::GetRefractionIndex(const int which) const
{
    assert(which<NumObjects);
    return Item[which]->GetRefractionIndex();
}

double Scene::IsReflective(const int which) const
{
    assert(which<NumObjects);
    return Item[which]->GetReflective();
}

double Scene::IsTransparent(const int which) const
{
    assert(which<NumObjects);
    return Item[which]->GetTransparent();
}

Colour Scene::GetSpecular(const int which) const
{
    assert(which<NumObjects);
    return Item[which]->GetSpecular();
}

double Scene::GetSpecExponent(const int which) const
{
        assert(which<NumObjects);
        return Item[which]->GetSpecExponent();
}


//////////////////////////////////////////////////////////////////////
// Methods to get the transform the last object which
// was added to the scene
//////////////////////////////////////////////////////////////////////

void Scene::rotateX(double angle)
{
	Item[NumObjects-1]->transMatrix.rotatex(angle);
}

void Scene::rotateY(double angle)
{
	Item[NumObjects-1]->transMatrix.rotatey(angle);
}

void Scene::rotateZ(double angle)
{
	Item[NumObjects-1]->transMatrix.rotatez(angle);
}

void Scene::scale(double sx, double sy, double sz)
{
	Item[NumObjects-1]->transMatrix.scale(sx, sy, sz);
}

void Scene::translate(double tx, double ty, double tz)
{
	Item[NumObjects-1]->transMatrix.translate(tx, ty, tz);
}

//////////////////////////////////////////////////////////////////////
// GetMatrix:  Returns the matrix of the last object added to 
// the scene.  Used for normal transformation.
//////////////////////////////////////////////////////////////////////
Matrix Scene::GetMatrix() const
{
	return Item[NumObjects-1]->transMatrix;
}

